/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2007,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifdef __PPU__

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

 
#define CPU_INFO	"/proc/cpuinfo"

/* get_cpu_frequency
 * -----------------
 * Parse /proc/cpuinfo for the CPU clock frequency. 
 * This information is returned in Hz. If the data 
 * can not be obtained, then 0.0 is returned.
 */

double get_cpu_frequency()
{
  FILE *fp;
  double freq = 0.0;
  char *line = NULL;
  size_t len = 0;
  ssize_t chrs_read;

  if ((fp = fopen(CPU_INFO, "r"))) {
    while ((chrs_read = getline(&line, &len, fp)) != -1) {
      if (sscanf(line, "clock : %lf", &freq) == 1) {
	if (strstr(line, "MHz")) {
	  freq *= 1.0e6;
	} else if (strstr(line, "GHz")) {
	  freq *= 1.0e9;
	}
	break;
      }
    }
    if (line) free(line);
    fclose(fp);
  }
  return (freq);
}

/* get_timebase_frequency
 * ----------------------
 * Parse /proc/cpuinfo for the timebase frequency. 
 * This information is returned in Hz. If the data 
 * can not be obtained, then 0.0 is returned.
 */

double get_timebase_frequency()
{
  FILE *fp;
  double freq = 0.0;
  char *line = NULL;
  size_t len = 0;
  ssize_t chrs_read;

  if ((fp = fopen(CPU_INFO, "r"))) {
    while ((chrs_read = getline(&line, &len, fp)) != -1) {
      if (sscanf(line, "timebase : %lf", &freq) == 1) {
	if (strstr(line, "KHz")) {
	  freq *= 1.0e3;
	} else if (strstr(line, "MHz")) {
	  freq *= 1.0e6;
	} else if (strstr(line, "GHz")) {
	  freq *= 1.0e9;
	}
	break;
      }
    }
    if (line) free(line);
    fclose(fp);
  }
  return (freq);
}

#endif /* __PPU__ */
