/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2007,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//#
//# ch_prof_flih.S
//# ------------------
//# This file contains the FLIH for a decrementer based profiler.
//# It consumes around 1/2 the cycles of a FLIH/SLIH pair and also
//# allows the interrupt to use a private interrupt stack.  The
//# private stack only needs to be large enough to backup volatile
//# registers.  If INTERRUPT_STACK_SIZE is set to 0 the runtime stack
//# will be used.  This requires that we save our working set of
//# registers at $SP+2000.  As $SP..$SP+2000 are reserved by the ABI
//# for leaf functions to use.                
//#
//# The interrupt handler is expected to be linked and located starting
//# at SPU local space address 0. It must be the first .o file sent to
//# the linker
//#
//# NOTE:       Even though this routine has been documented to be
//#             fast, it has not been optimized with respect to
//#             instruction scheduling so that readabiliy of the
//#             code is improved.
//#     
//# This function can be assembled to use either the standard runtime
//# stack or a private interrupt stack for temporary state variables.
//# Setting INTERRUPT_STACK_SIZE to a non-zero value will result in 
//# code the maintains a separate stack for the interrupt handler.
//# The size of the separate interrupt stack is defined to be
//# its value in bytes.
//#
        
.set            INTERRUPT_STACK_SIZE, 224

.set            DECR_COUNT, 10000       
        
.extern         ch_prof_tagid_table // Counts for each tag
.extern         ch_prof_ch_table    // Counts for each channel
.extern         ch_prof_ticks       // Number of decrementer counts

.extern         ch_prof_rdch_cntr   // Number of read stalls
.extern         ch_prof_wrch_cntr   // Number of write stalls
.extern         ch_prof_misc_cntr   // Non stalls

        .section .interrupt,"ax"
        .align          3
        .global spu_handler_fast
spu_handler_fast:               
        
.if INTERRUPT_STACK_SIZE
        //# Establish a private stack used only by the interrupt handler.
        //# This interrupt handler will not support nested interrupts.
        
        .set            STACK_SKIP, 0
        stqr            $SP, main_stack_ptr
        ila             $SP, interrupt_stack-16
.else
        //# Adjust the stack pointer to skip the maximum register save area
        //# (STACK_SKIP quadword registers) in case an interrupt occurred while
        //# executing a leaf function that used the stack area without actually 
        //# allocating its own stack frame.
        
        .set            STACK_SKIP, 125
.endif
        
        //# Save the working registers used by the interrupt handler
        //# r3   Scratch
        //# r4   Scratch
        //# r5   Scratch
        //# r6   Event mask
        //# r7   The instructioon (*SER0)
        //# r8   scratch (counters)
        //# r9   ch_prof_rdch?
        //# r10  ch_prof_wrch?
        //# r11  ch_prof_misc?
        //# r12  scratch
        //# r13  scratch
        //# r14  scratch
        //# r15  Channel id
        stqd            $3, -(STACK_SKIP+0)*16($SP)
        stqd            $4, -(STACK_SKIP+1)*16($SP)
        stqd            $5, -(STACK_SKIP+2)*16($SP)
        stqd            $6, -(STACK_SKIP+3)*16($SP)
        stqd            $7, -(STACK_SKIP+4)*16($SP)
        stqd            $8, -(STACK_SKIP+5)*16($SP)
        stqd            $9, -(STACK_SKIP+6)*16($SP)
        stqd            $10, -(STACK_SKIP+7)*16($SP)
        stqd            $11, -(STACK_SKIP+8)*16($SP)
        stqd            $12, -(STACK_SKIP+9)*16($SP)
        stqd            $13, -(STACK_SKIP+10)*16($SP)
        stqd            $14, -(STACK_SKIP+11)*16($SP)
        stqd            $15, -(STACK_SKIP+12)*16($SP)
        //# Update INTERRUPT_STACK_SIZE if you add any more registers
                
        rdch            $3, $SPU_RdEventStat            //# read event status
        rdch            $6, $SPU_RdEventMask            //# read event mask

        andc            $4,$6,$3                        //# dissable events
        wrch            $SPU_WrEventMask, $4
        wrch            $SPU_WrEventAck, $3             //# acknowledge the event

        //# Process the event. For this example, we only support a single
        //# event, the decrementer event. Halt if we get something unexpected.
        ceqi            $3, $3, 0x20
        heqi            $3, 0
        

        //# Increment the ch_prof count
        lqr             $3, ch_prof_ticks
        ila             $5, ch_prof_ticks
        ai              $4, $3, 1
        cwd             $5, 0($5)
        shufb           $3, $4, $3, $5
        stqr            $3, ch_prof_ticks

        //# Get the instruction at SRR0
        rdch            $3, $SPU_RdSRR0                 //# The PC at interrupt time
        lqd             $4, 0($3)                       //# The actual instruction *PC
        rotqby          $7, $4, $3                      //# Move the load to word 0
        lqa             $5, shuffle_splat_word0_to_word0_3
        shufb           $7, $7, $7, $5                 //# splat the instruction across $r7
        
        //# Disassemble the instruction at SRR0
        ilh             $3, 0x0
        ilhu            $3, 0xfff0                     //# Instruction mask 0xfff00000
        and             $4, $7, $3

        ilhu            $3, 0x01a0                     //# Read channel instruction?
        ceq             $9, $4, $3
        
        ilhu            $3, 0x21a0                     //# write channel instruction?
        ceq             $10, $4, $3

        or              $11, $9, $10                   //# Neither read nor write
        ceqi            $11, $11, 0
                
        //# Increment the correct counts
        lqr             $8, ch_prof_rdch_cntr
        ai              $3, $8, 1
        selb            $4, $8, $3, $9
        ila             $5, ch_prof_rdch_cntr
        cwd             $5, 0($5)
        shufb           $3, $4, $8, $5
        stqr            $3, ch_prof_rdch_cntr
                
        lqr             $8, ch_prof_wrch_cntr
        ai              $3, $8, 1
        selb            $4, $8, $3, $10
        ila             $5, ch_prof_wrch_cntr
        cwd             $5, 0($5)
        shufb           $3, $4, $8, $5
        stqr            $3, ch_prof_wrch_cntr

        lqr             $8, ch_prof_misc_cntr
        ai              $3, $8, 1
        selb            $4, $8, $3, $11
        ila             $5, ch_prof_misc_cntr
        cwd             $5, 0($5)
        shufb           $3, $4, $8, $5
        stqr            $3, ch_prof_misc_cntr

        //# Increment the channel count
        rotqmbii        $3, $7, -7                      //# Get the channel number
        andi            $15, $3, 0x1f
        shli            $3, $15, 2                      //# indexed for sizeof(int)
        ila             $4, ch_prof_ch_table
        a               $5, $3, $4                      //# &ch_prof_ch_table[ch]
        lqx             $12, $3, $4                     //# ch_prof_ch_table[ch]
        cwx             $8, $3, $4
        ori             $13, $12, 0
        rotqby          $12, $12, $5
        ai              $14, $12, 1
        selb            $12, $14, $12, $11              //# if !(rdch||wrch) count++
        shufb           $13, $12, $13, $8
        stqx            $13, $3, $4

        //# Increment the tagmask count
        rdch            $3, $MFC_RdTagMask              //# read the tag mask
        clz             $4, $3
        sfi             $3, $4, 31                      //# which tag
        shli            $3, $3, 2                       //# indexed for sizeof(int)
        ila             $4, ch_prof_tagid_table
        a               $5, $3, $4                      //# &ch_prof_ch_table[ch]
        lqx             $12, $3, $4                     //# ch_prof_ch_table[ch]
        cwx             $8, $3, $4
        ori             $13, $12, 0
        rotqby          $12, $12, $5
        ai              $14, $12, 1
        selb            $14, $12, $14, $9               //# if (rdch)
        ceqi            $5, $15, 24
        selb            $12, $12, $14, $5               //# && channel == 24
        shufb           $13, $12, $13, $8
        stqx            $13, $3, $4

        //# reset the decrementer counter       
        ila             $5, DECR_COUNT
        wrch            $SPU_WrDec, $5
        
        //# Restore event mask and all working registers
        wrch            $SPU_WrEventMask, $6            //# restore event mask
        lqd             $3, -(STACK_SKIP+0)*16($SP)
        lqd             $4, -(STACK_SKIP+1)*16($SP)
        lqd             $5, -(STACK_SKIP+2)*16($SP)
        lqd             $6, -(STACK_SKIP+3)*16($SP)
        lqd             $7, -(STACK_SKIP+4)*16($SP)
        lqd             $8, -(STACK_SKIP+5)*16($SP)
        lqd             $9, -(STACK_SKIP+6)*16($SP)
        lqd             $10, -(STACK_SKIP+7)*16($SP)
        lqd             $11, -(STACK_SKIP+8)*16($SP)
        lqd             $12, -(STACK_SKIP+9)*16($SP)
        lqd             $13, -(STACK_SKIP+10)*16($SP)
        lqd             $14, -(STACK_SKIP+11)*16($SP)
        lqd             $15, -(STACK_SKIP+12)*16($SP)

.if INTERRUPT_STACK_SIZE
        lqr             $SP, main_stack_ptr
.endif
        
        //# Return from interrupt and re-enable interrupts
        irete

.if     INTERRUPT_STACK_SIZE
        //# Private interrupt stack
        .skip           INTERRUPT_STACK_SIZE
        .align          4
  interrupt_stack:                      //# top of interrupt stack
        

        .align          4
  shuffle_splat_word0_to_word0_3:
        .octa           0x00010203000102030001020300010203
        
        //# Place to save main runtime stack pointer
  main_stack_ptr:
        .skip           16

.else   //# !INTERRUPT_STACK_SIZE

        .align          4
  shuffle_splat_word0_to_word0_3:
        .octa           0x00010203000102030001020300010203
.endif  //# INTERRUPT_STACK_SIZE
