/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <libspe2.h>
#include <pthread.h>

extern spe_program_handle_t race_check_example_spu;
extern spe_program_handle_t race_check_example_fixed_spu;

unsigned char buffer[65536] __attribute__ ((aligned(128)));


void *ppu_pthread_function(void *arg) {
  spe_context_ptr_t ctx;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  
  ctx = *((spe_context_ptr_t *)arg);
  if (spe_context_run(ctx, &entry, 0, NULL, NULL, NULL) < 0) {
    perror ("Failed running context");
    exit (1);
  }
  pthread_exit(NULL);
}


static struct option longopts[] = {
    /* name     has_arg      flag   val */
    { "help",   no_argument, NULL,  'h'},
    { "fixed",  no_argument, NULL,  'f'},
    {0, 0, 0, 0}
};




int main(int argc, char *argv[])
{
  int c;
  int fixed = 0;
  spe_context_ptr_t ctx;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  
 
  while ((c = getopt_long(argc, argv, "hf", longopts, NULL)) != -1) {
    switch (c) {
    case 'f':		/* fix reace conditions */
      fixed = 1;
      break;

    case 'h':		/* help */
    default:
      printf("USAGE: %s [-f]\n", argv[0]);
      printf("       where: -f specifies that the fixed version of the race condition example is to be run.\n");
      return 1;
      break;
    }
  }

  /* Create context */
  if ((ctx = spe_context_create (0, NULL)) == NULL) {
    perror ("Failed creating context");
    exit (1);
  }

  /* Load program into context */
  if (spe_program_load (ctx, (fixed) ? &race_check_example_fixed_spu : &race_check_example_spu)) {
    perror ("Failed loading program");
    exit (1);
  }

  /* Execute the SPE example program */
  if (spe_context_run(ctx, &entry, 0, buffer, NULL, NULL) < 0) {
    perror ("Failed running context");
    exit (1);
  }

  /* Destroy context */
  if (spe_context_destroy (ctx) != 0) {
    perror("Failed destroying context");
    exit (1);
  }

  return (0);
}
