/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2007,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//#
//# spu_handler_fast.s
//# ------------------
//# This file contains a example fast SPU interrupt handler. This
//# handler is meant to serve as example code the demonstrates how
//# one can implement a low latency interrupt handler. To do this,
//# the interrupt handler must be written entirely in assembly. This 
//# implementation assumes that no external calls are made.
//#
//# The interrupt handler is expected to be linked and located starting
//# at SPU local space address 0. This is accomplished by specify
//# the section to be .interrupt.
//#
//# NOTE:	Even though this routine has been documented to be
//#		fast, it has not been optimized with respect to
//#		instruction scheduling so that readabiliy of the
//#		code is improved.
//#	
//# This function can be assembled to use either the standard runtime
//# stack or a private interrupt stack for temporary state variables.
//# Setting INTERRUPT_STACK_SIZE to a non-zero value will result in 
//# code the maintains a separate stack for the interrupt handler.
//# The size of the separate interrupt stack is defined to be
//# its value in bytes.
//#
	
.set		INTERRUPT_STACK_SIZE, 0

.set		DECR_COUNT, 10000	
	
.extern		event_count	

	.section .interrupt,"ax"
	.align		3
	.global	spu_handler_fast
spu_handler_fast:		
	
.if INTERRUPT_STACK_SIZE
	//# Establish a private stack used only by the interrupt handler.
	//# This interrupt handler will not support nested interrupts.
	
	.set		STACK_SKIP, 0
	stqr		$SP, main_stack_ptr
	ila		$SP, interrupt_stack-16
.else
	//# Adjust the stack pointer to skip the maximum register save area
	//# (STACK_SKIP quadword registers) in case an interrupt occurred while
	//# executing a leaf function that used the stack area without actually 
	//# allocating its own stack frame.
	
	.set		STACK_SKIP, 125
.endif
	
	//# Save the working registers used by the interrupt handler
	stqd		$3, -(STACK_SKIP+0)*16($SP)
	stqd		$4, -(STACK_SKIP+1)*16($SP)
	stqd		$5, -(STACK_SKIP+2)*16($SP)
	stqd		$6, -(STACK_SKIP+3)*16($SP)
		
	
	rdch		$3, $SPU_RdEventStat		//# read event status
	rdch		$6, $SPU_RdEventMask		//# read event mask

	andc		$4,$6,$3			//# dissable events
	wrch		$SPU_WrEventMask, $4
	wrch		$SPU_WrEventAck, $3		//# acknowledge the event

	//# Process the event. For this example, we only support a single
	//# event, the decrementer event. Halt if we get something unexpected.
	ceqi		$3, $3, 0x20
	heqi		$3, 0
	

	lqr		$3, event_count			//# increment the global event count
	ai		$4, $3, 1
	ila		$5, event_count
	cwd		$5, 0($5)
	shufb		$3, $4, $3, $5
	stqr		$3, event_count

	ila		$5, DECR_COUNT			//# reset the decrementer counter 	
	wrch		$SPU_WrDec, $5
	
	// Restore event mask and all working registers
	wrch		$SPU_WrEventMask, $6		//# restore event mask
	lqd		$3, -(STACK_SKIP+0)*16($SP)
	lqd		$4, -(STACK_SKIP+1)*16($SP)
	lqd		$5, -(STACK_SKIP+2)*16($SP)
	lqd		$6, -(STACK_SKIP+3)*16($SP)

.if	INTERRUPT_STACK_SIZE	
	lqr		$SP, main_stack_ptr
.endif
        	
#ifdef SPU_BISLED_H
	//# Return from SPU bisled event handler using link register
	bi		$0
#else // normal use for spu_flih
	//# Return from interrupt and re-enable interrupts
	irete
#endif // SPU_BISLED_H
	
.if	INTERRUPT_STACK_SIZE
	//# Private interrupt stack
	.skip		INTERRUPT_STACK_SIZE
        .align          4
  interrupt_stack:			//# top of interrupt stack
	
	//# Place to save main runtime stack pointer
  main_stack_ptr:
	.skip		16
.endif
