/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <stdio.h>
#include <spu_slih_reg.h>
#include <spu_intrinsics.h>
#include <spu_mfcio.h>

/* Number of timebase ticks per decrementer event. This number
 * must be large enough that we don't spend all our time in the
 * interrupt handler.
 */
#define DECR_COUNT	10000


volatile unsigned int event_count = 0;


unsigned int decr_handler(unsigned int status)
{
  status &= ~MFC_DECREMENTER_EVENT;
  event_count++;

  /* Reset counter.
   */
  spu_writech(SPU_WrDec, DECR_COUNT);
  return (status);
}


int main(void)
{
  /* Register a decrementer event handler
   */
  spu_slih_reg(MFC_DECREMENTER_EVENT, decr_handler);

  /* Initialize the decrementer state including:
   * 1) event mask
   * 2) decrementer counter
   * 3) interrupt enable state
   */
  spu_writech(SPU_WrEventMask, MFC_DECREMENTER_EVENT);
  spu_writech(SPU_WrDec, DECR_COUNT);

  spu_ienable();

  /* Wait for 10 decrementer event interrupts..
   */
  while (event_count < 10);

  printf("Test successfully completed.\n");
  return (0);
}
